@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -195,8 +195,10 @@ gboolean at_util_parse_reg_unsolicited(GAtResult *result, const char *prefix,
 
 	switch (vendor) {
 	case OFONO_VENDOR_GOBI:
+	case OFONO_VENDOR_ZTE:
 	case OFONO_VENDOR_HUAWEI:
 	case OFONO_VENDOR_NOVATEL:
+	case OFONO_VENDOR_SPEEDUP:
 		if (g_at_result_iter_next_unquoted_string(&iter, &str) == TRUE)
 			l = strtol(str, NULL, 16);
 		else
@@ -257,8 +259,10 @@ gboolean at_util_parse_reg(GAtResult *result, const char *prefix,
 
 		/* Sometimes we get an unsolicited CREG/CGREG here, skip it */
 		switch (vendor) {
+		case OFONO_VENDOR_ZTE:
 		case OFONO_VENDOR_HUAWEI:
 		case OFONO_VENDOR_NOVATEL:
+		case OFONO_VENDOR_SPEEDUP:
 			r = g_at_result_iter_next_unquoted_string(&iter, &str);
 
 			if (r == FALSE || strlen(str) != 1)
@@ -280,8 +284,10 @@ gboolean at_util_parse_reg(GAtResult *result, const char *prefix,
 
 		switch (vendor) {
 		case OFONO_VENDOR_GOBI:
+		case OFONO_VENDOR_ZTE:
 		case OFONO_VENDOR_HUAWEI:
 		case OFONO_VENDOR_NOVATEL:
+		case OFONO_VENDOR_SPEEDUP:
 			r = g_at_result_iter_next_unquoted_string(&iter, &str);
 
 			if (r == TRUE)
@@ -581,6 +587,9 @@ struct at_util_sim_state_query *at_util_sim_state_query_new(GAtChat *chat,
 
 void at_util_sim_state_query_free(struct at_util_sim_state_query *req)
 {
+	if (req == NULL)
+		return;
+
 	if (req->cpin_poll_source > 0)
 		g_source_remove(req->cpin_poll_source);
 
