@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -309,6 +309,19 @@ static void service_number_free(struct service_number *num)
 	g_free(num);
 }
 
+static void call_state_watches(struct ofono_sim *sim)
+{
+	GSList *l;
+	ofono_sim_state_event_cb_t notify;
+
+	for (l = sim->state_watches->items; l; l = l->next) {
+		struct ofono_watchlist_item *item = l->data;
+		notify = item->notify;
+
+		notify(sim->state, item->notify_data);
+	}
+}
+
 static DBusMessage *sim_get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -806,6 +819,9 @@ static DBusMessage *sim_enter_pin(DBusConnection *conn, DBusMessage *msg,
 	if (type == OFONO_SIM_PASSWORD_NONE || type != sim->pin_type)
 		return __ofono_error_invalid_format(msg);
 
+	if (password_is_pin(type) == FALSE)
+		return __ofono_error_invalid_format(msg);
+
 	if (!__ofono_is_valid_sim_pin(pin, type))
 		return __ofono_error_invalid_format(msg);
 
@@ -945,18 +961,15 @@ static void sim_get_image(struct ofono_sim *sim, unsigned char id,
 	unsigned short iidf_offset;
 	unsigned short iidf_len;
 
-	image = sim_fs_get_cached_image(sim->simfs, id);
-
-	if (image != NULL) {
-		sim_get_image_cb(sim, id, image, FALSE);
-		goto watch;
-	}
-
-	if (sim->efimg_length <= (id * 9)) {
+	if (sim->efimg_length <= id * 9) {
 		sim_get_image_cb(sim, id, NULL, FALSE);
 		return;
 	}
 
+	image = sim_fs_get_cached_image(sim->simfs, id);
+	if (image != NULL)
+		sim_get_image_cb(sim, id, image, FALSE);
+
 	efimg = &sim->efimg[id * 9];
 
 	iidf_id = efimg[3] << 8 | efimg[4];
@@ -964,12 +977,9 @@ static void sim_get_image(struct ofono_sim *sim, unsigned char id,
 	iidf_len = efimg[7] << 8 | efimg[8];
 
 	/* read the image data */
-	ofono_sim_read_bytes(sim->context, iidf_id, iidf_offset, iidf_len,
-				sim_iidf_read_cb, sim);
-
-watch:
-	if (sim->efimg_length <= id * 9)
-		return;
+	if (image == NULL)
+		ofono_sim_read_bytes(sim->context, iidf_id, iidf_offset,
+					iidf_len, sim_iidf_read_cb, sim);
 
 	if (sim->iidf_watch_ids[id] > 0)
 		return;
@@ -1361,25 +1371,18 @@ static void sim_ready(enum ofono_sim_state new_state, void *user)
 
 static void sim_set_ready(struct ofono_sim *sim)
 {
-	GSList *l;
-	ofono_sim_state_event_cb_t notify;
-
 	if (sim == NULL)
 		return;
 
-	if (sim->state != OFONO_SIM_STATE_INSERTED)
+	if (sim->state != OFONO_SIM_STATE_INSERTED &&
+			sim->state != OFONO_SIM_STATE_LOCKED_OUT)
 		return;
 
 	sim->state = OFONO_SIM_STATE_READY;
 
 	sim_fs_check_version(sim->simfs);
 
-	for (l = sim->state_watches->items; l; l = l->next) {
-		struct ofono_watchlist_item *item = l->data;
-		notify = item->notify;
-
-		notify(sim->state, item->notify_data);
-	}
+	call_state_watches(sim);
 }
 
 static void sim_imsi_cb(const struct ofono_error *error, const char *imsi,
@@ -1729,6 +1732,8 @@ static void sim_efphase_read_cb(int ok, int length, int record,
 
 static void sim_initialize_after_pin(struct ofono_sim *sim)
 {
+	sim->context = ofono_sim_context_create(sim);
+
 	ofono_sim_read(sim->context, SIM_EFPHASE_FILEID,
 			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
 			sim_efphase_read_cb, sim);
@@ -1930,12 +1935,8 @@ skip_efpl:
 						&sim->language_prefs);
 
 	/* Proceed with sim initialization if we're not merely updating */
-	if (!sim->language_prefs_update) {
-		if (sim->context == NULL)
-			sim->context = ofono_sim_context_create(sim);
-
+	if (!sim->language_prefs_update)
 		__ofono_sim_recheck_pin(sim);
-	}
 
 	sim->language_prefs_update = FALSE;
 }
@@ -2178,6 +2179,12 @@ ofono_bool_t __ofono_sim_service_available(struct ofono_sim *sim,
 	return FALSE;
 }
 
+ofono_bool_t __ofono_sim_cphs_service_available(struct ofono_sim *sim,
+						int cphs_service)
+{
+	return sim_cphs_is_active(sim->cphs_service_table, cphs_service);
+}
+
 static void sim_inserted_update(struct ofono_sim *sim)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -2216,6 +2223,13 @@ static void sim_free_early_state(struct ofono_sim *sim)
 
 static void sim_free_main_state(struct ofono_sim *sim)
 {
+	int i;
+
+	for (i = 0; i < OFONO_SIM_PASSWORD_INVALID; i++)
+		sim->pin_retries[i] = -1;
+
+	memset(sim->locked_pins, 0, sizeof(sim->locked_pins));
+
 	if (sim->imsi) {
 		g_free(sim->imsi);
 		sim->imsi = NULL;
@@ -2256,7 +2270,10 @@ static void sim_free_main_state(struct ofono_sim *sim)
 		sim->efsst_length = 0;
 	}
 
+	sim->phase = OFONO_SIM_PHASE_UNKNOWN;
+	sim->cphs_phase = OFONO_SIM_CPHS_PHASE_NONE;
 	sim->mnc_length = 0;
+	memset(sim->cphs_service_table, 0, 2);
 
 	if (sim->efimg) {
 		g_free(sim->efimg);
@@ -2286,9 +2303,6 @@ static void sim_free_state(struct ofono_sim *sim)
 
 void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 {
-	ofono_sim_state_event_cb_t notify;
-	GSList *l;
-
 	if (inserted == TRUE && sim->state == OFONO_SIM_STATE_NOT_PRESENT)
 		sim->state = OFONO_SIM_STATE_INSERTED;
 	else if (inserted == FALSE && sim->state != OFONO_SIM_STATE_NOT_PRESENT)
@@ -2300,13 +2314,7 @@ void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 		return;
 
 	sim_inserted_update(sim);
-
-	for (l = sim->state_watches->items; l; l = l->next) {
-		struct ofono_watchlist_item *item = l->data;
-		notify = item->notify;
-
-		notify(sim->state, item->notify_data);
-	}
+	call_state_watches(sim);
 
 	if (inserted)
 		sim_initialize(sim);
@@ -2392,9 +2400,8 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 			/* Force the sim state out of READY */
 			sim_free_main_state(sim);
 
-			sim->state = OFONO_SIM_STATE_INSERTED;
-			__ofono_modem_sim_reset(
-					__ofono_atom_get_modem(sim->atom));
+			sim->state = OFONO_SIM_STATE_LOCKED_OUT;
+			call_state_watches(sim);
 		}
 		break;
 	}
